/* Copyright 2003 Heikki Orsila <heikki.orsila@ee.tut.fi>
   Copyright 2003 Harry Sintonen
*/

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#include "amiga-shell.h"

#include "uade-os.h"
#include "../osdep/strl.c"

#include <exec/types.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dosasl.h>
#include <proto/dos.h>


void as_setup(void (*interrupt)(void));
void as_enqueue_song(char *file);
void as_get_next(char *dst, int maxsize);
int as_song_list_empty(void);


struct songnode_t {
  struct songnode_t *next;
  char filename[PATH_MAX];
};

static struct songnode_t *SongList = 0;

static void (*as_interrupt)(void) = 0;

/* catch CTRL-C and cause REBOOT */
static void as_sigint_handler(int signum) {
  signum = signum;
  if (as_interrupt) {
    as_interrupt();
  } else {
    fprintf(stderr, "uade: non-initialized sigint interrupt\n");
  }
}

static int as_cleanup_installed = 0;
static BPTR as_stdin = NULL;
static int as_stdin_does_cooked = 0;
static int as_stdin_is_cooked = 0;

static
void as_cleanup(void) {
  if (as_stdin) {
    if (as_stdin_is_cooked) {
      as_stdin_is_cooked = 0;
      /* back to con mode */
      SetMode(as_stdin, 0);
    }
    as_stdin = NULL;
  }
}

void as_setup(void (*interrupt)(void)) {
  if (!as_cleanup_installed) {
    as_cleanup_installed = 1;
    if (atexit(as_cleanup) < 0) {
      uade_exit(-1);
    }
  }

  as_stdin = Input();

  if (!as_stdin_does_cooked) {
    /* only do console interaction if stdin is interactive */
    if (IsInteractive(as_stdin)) {
      /* enter RAW 'cooked' mode */
      if (!SetMode(as_stdin, 1)) {
        printf("as_setup: can't put console to cooked raw mode!\n");
        uade_exit(-1);
      }
      /* back to con mode */
      SetMode(as_stdin, 0);
      /* yeah, it does cooked mode */
      as_stdin_does_cooked = 1;
    }
  }

  as_interrupt = interrupt;
  /* create CTRL-C break for cmdline tool */
  uade_create_signalhandler(as_sigint_handler, UADE_SIGINT);
}


static
void enqueue_song(char *file) {
  struct songnode_t *s = SongList;
  struct songnode_t *new;

  /* allow queuing of zero length string */
  if (uade_is_regular_file(file) || strlen(file) == 0) {
    new = malloc(sizeof(struct songnode_t));
    new->next = 0;
    strlcpy(new->filename, file, sizeof(new->filename));
    if (s) {
      while(s->next) s = s->next;
      s->next = new;
    } else {
      SongList = new;
    }
  }
}


struct anchordata {
  struct AnchorPath ap;
  char buf[1024];
};

void as_enqueue_song(char *file) {
  char _data[sizeof(struct anchordata) + 3];
  struct anchordata *data = (void *) (((ULONG) _data + 3) & ~3);
  LONG err, count = 0;

  data->ap.ap_BreakBits  = SIGBREAKF_CTRL_C;
  data->ap.ap_FoundBreak = 0;
  data->ap.ap_Flags      = 0;
  data->ap.ap_Strlen     = sizeof(data->buf);

  for (err = MatchFirst(file, &data->ap); !err; err = MatchNext(&data->ap)) {
    enqueue_song(data->ap.ap_Buf);
    count++;
  }

  /* If nothing matched, assume it's literal name */
  if (count == 0 && err == ERROR_NO_MORE_ENTRIES) {
    enqueue_song(file);
  }
}

int as_song_list_empty(void) {
  return SongList ? 0 : 1;
}

void as_get_next(char *dst, int maxsize) {
  char playername[PATH_MAX];
  char tempname[PATH_MAX];

  if (!SongList)
    return;

  strlcpy(dst, SongList->filename, maxsize);
  SongList = SongList->next;
}

static int as_interaction(struct uade_command *cmd) {
  LONG c;

  if (!as_stdin_does_cooked) {
    goto no_interaction;
  }
  
  /* enter RAW 'cooked' mode */
  if (!as_stdin_is_cooked) {
    if (!SetMode(as_stdin, 1)) {
      goto no_interaction;
    }

    as_stdin_is_cooked = 1;

    /* flush possible pending chars in the buffer */
    Flush(as_stdin);
  }

  /* timeout 0 == poll */
  if (WaitForChar(as_stdin, 0)) {

    c = FGetC(as_stdin);

    switch (c)
    {
      case -1:
	goto no_interaction;

      case 0x03: /* CTRL-C */
        as_sigint_handler(1);
        break;

      case 0x9b:
	{
	  /* read esc sequence (fkeys, other special keys) */
	  char seq[12];
	  int i = 0;
	  do {
	    c = FGetC(as_stdin);
	    seq[i++] = c;
	  } while (c != -1 && c != '~' && i < (sizeof(seq) - 1));
	  seq[i] = '\0';
	  break;
	}
      case '\n':
      case '\r':
	break;

      case 'n':
	cmd->type = UADE_REBOOT;
	return 1;

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	cmd->type = UADE_SETSUBSONG;
	cmd->ret = malloc(sizeof(int));
	if (!cmd->ret) {
	  goto no_interaction;
	}
	*((int *) cmd->ret) = (int) (c - '0');
	return 1;

      default:
	fprintf(stderr, "uade: unknown command for interaction\n");
	break;
    }
  }

  return 0;

 no_interaction:
  cmd->type = UADE_NO_INTERACTION;
  return 1;
}

void as_functions(struct uade_local *local) {
  local->setup = as_setup;
  local->enqueue_song = as_enqueue_song;
  local->get_next = as_get_next;
  local->interaction = as_interaction;
  local->song_list_empty = as_song_list_empty;
};
