/* UADE - Unix Amiga Delitracker Emulator
 * Copyright 2000-2003, Heikki Orsila <heikki.orsila@ee.tut.fi>
 * See http://ee.tut.fi/~heikki/uade.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h>
#include <netinet/in.h>
#include <limits.h>
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "events.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "debug.h"
#include "gensound.h"

#include "uade.h"
#include "uade-os.h"
#include "../config.h"

#ifdef HAVE_AMIGA_SHELL
#include "amiga-shell.h"
#endif
#ifdef HAVE_UNIX_SHELL
#include "unix-shell.h"
#endif
#ifdef HAVE_XMMS_SLAVE
#include "xmms-slave.h"
#endif

#include "../decrunch/decrunch.h"
#include "../plugindir/amifilemagic.h"

#include "../osdep/strl.c"

static void check_name_extension(char *playername, char *modname, int maxl);
static int uade_calc_reloc_size(unsigned int *src);
static void uade_flush_sound(void);
static int uade_get_long(int addr);
static void uade_print_help(int problemcode);
static void uade_put_long(int addr,int val);
static void uade_reset_counters(void);
static int uade_safe_load(int dst, FILE *file, int maxlen);

static const int SCORE_MODULE_ADDR   = 0x100;
static const int SCORE_MODULE_LEN    = 0x104;
static const int SCORE_PLAYER_ADDR   = 0x108;
static const int SCORE_RELOC_ADDR    = 0x10C;
static const int SCORE_USER_STACK    = 0x110;
static const int SCORE_SUPER_STACK   = 0x114;
static const int SCORE_FORCE         = 0x118;
static const int SCORE_SET_SUBSONG   = 0x11c;
static const int SCORE_SUBSONG       = 0x120;
static const int SCORE_NTSC          = 0x124;
static const int SCORE_MODULE_NAME_ADDR = 0x128;
static const int SCORE_HAVE_SONGEND  = 0x12C;
static const int SCORE_POSTPAUSE     = 0x180;
static const int SCORE_PREPAUSE      = 0x184;
static const int SCORE_DELIMON       = 0x188;
static const int SCORE_EXEC_DEBUG    = 0x18C;
static const int SCORE_VOLUME_TEST   = 0x190;
static const int SCORE_DMA_WAIT      = 0x194;
static const int SCORE_MODULECHANGE  = 0x198;

static const int SCORE_INPUT_MSG     = 0x200;
static const int SCORE_MIN_SUBSONG   = 0x204;
static const int SCORE_MAX_SUBSONG   = 0x208;
static const int SCORE_CUR_SUBSONG   = 0x20C;

static const int SCORE_OUTPUT_MSG    = 0x300;

static int uade_highmem = 0x200000;

/* is uade commanded by external application? */
int uade_remote_slave = 0;

static struct uade_local local;
static struct uade_remote remote;

struct uade_remote *uade_remote = 0; /* if slave is used, this is &remote */

int uade_slave_end_reported = 0;

struct uade_song uade_song;

char score_playername[256];      /* playername that eagleplayer reports */
char score_modulename[256];      /* modulename that eagleplayer reports */
char score_formatname[256];	 /* formatname that eagleplayer reports */

static int uade_execdebugboolean = 0;
static int uade_repeat = 0;

static int uade_max_subsong;
static int uade_min_subsong;
static int uade_cur_subsong;
static int uade_one_subsong = 0;

/* don't touch */
static int uade_dmawait = 0;
static int voltestboolean = 0;
static int disable_modulechange = 0;

static int playerforced = 0;

/* output sound data to pipe */
int uade_usingoutpipe = 0;
/* This should be non-zero if native sound output shouldn't be used.
   This is set to non-zero in slavemode and outpipemode              */
int uade_local_sound = 1;
int uade_swap_output_bytes = 0;

int uade_reboot;

int uade_debug = 0;

int uade_do_panning = 0;
float uade_pan_value = 1.0f;

/* contains uade's command line name */
static char *uadecmdlinename = 0;

static int uade_vsync_counter;
static int uade_audxdat_counter;
static int uade_audxvol_counter;
static int uade_audxlch_counter;

int uade_zero_sample_count;

int uade_timeout = -1;		/* default timeout infinite */
int uade_subsong_timeout = -1;	/* default per subsong timeout infinite */
int uade_silence_timeout = 20;	/* default silence timeout */

int uade_time_critical = 0;
static int uade_speed_hack = 0;

int uade_interactive = 0;

static void dummy_enqueue_song(char *file) {
  fprintf(stderr, "uade: bug: dummy_enqueue_song() called with %s\n", file);
}
static int dummy_song_list_empty(void) {
  fprintf(stderr, "uade: bug: dummy_song_list_empty() called\n");
  return 1;
}
static void dummy_get_next(char *dst, int maxsize) {
  dst = dst; maxsize = maxsize;
  fprintf(stderr, "uade: bug: dummy_get_next() called\n");
}
static void dummy_setup(void (*interrupt)(void)) {interrupt = interrupt;}


static void uade_interrupt(void) {
  if (!uade_debug) {
    int ltime;
    ltime = uade_timer_poll();
    if (uade_reboot || (ltime > 0 && ltime < 100)) {
      fprintf(stderr,"uade: exit forced with ctrl-c\n");
      uade_exit(0);
    }
    uade_reboot = 1;
  } else {
    activate_debugger();
  }
}

void uade_option(int argc, char **argv) {
  int i;
  char **slaveargs = 0;

  memset(&local, 0, sizeof(local));

  local.setup = dummy_setup;
  local.enqueue_song = dummy_enqueue_song;
  local.get_next = dummy_get_next;
  local.song_list_empty = dummy_song_list_empty;

#ifdef HAVE_UNIX_SHELL
  us_functions(&local);
#endif

#ifdef HAVE_AMIGA_SHELL
  as_functions(&local);
#endif

  uadecmdlinename = strdup(argv[0]);

  uade_song.playername[0] = 0;
  uade_song.song_end_possible = 1;

  uade_get_path(uade_song.scorename, UADE_PATH_SCORE, sizeof(uade_song.scorename));

  for (i = 1; i < argc;) {

    /* check for playername parameter */
    if (strcasecmp(argv[i], "-P") == 0) {
      if ((i+1) < argc) {
	strlcpy(uade_song.playername, argv[i+1], sizeof(uade_song.playername));
	playerforced = 1;
	i+=2;
      } else {
	fprintf(stderr, "Player name missing for -P parameter\n\n");
	uade_print_help(1);
	uade_exit(-1);
      }

      /* check for modulename parameter */
    } else if (strcasecmp(argv[i], "-M") == 0) {

      if ((i+1) < argc) {
	local.enqueue_song(argv[i+1]);
	i+=2;
      } else {
	fprintf(stderr, "Song name missing for -M parameter\n\n");
	uade_print_help(1);
	uade_exit(-1);
      }

      /* check for scorename parameter */
    } else if (strcasecmp(argv[i], "-S") == 0) {
      if ((i+1) < argc) {
	strlcpy(uade_song.scorename, argv[i+1], sizeof(uade_song.scorename));
      } else {
	fprintf(stderr, "missing parameter for -S\n\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      i+=2;

    } else if (strcmp(argv[i], "-slave") == 0) {
      if ((i+1) < argc) {
	slaveargs = &argv[i+1];
	uade_remote_slave = 1;
	uade_local_sound = 0;
      } else {
	fprintf(stderr, "parameter missing for -slave\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      i+=2;

    } else if (strcmp(argv[i], "-outpipe") == 0) {
      if ((i+1) < argc) {
	uade_usingoutpipe = 1;
	uade_local_sound = 0;
	uade_init_outpipe(argv[i+1]);
      } else {
	fprintf(stderr, "parameter missing for -outpipe\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      i+=2;

    } else if (strcmp(argv[i], "-force") == 0) {
      uade_song.force_by_default = 1;
      i++;

    } else if (strcmp(argv[i], "-execdebug") == 0) {
      fprintf(stderr, "-execdebug parameter\n");
      uade_execdebugboolean = 1;
      i++;

    } else if (strcmp(argv[i], "-voltest") == 0) {
      fprintf(stderr, "-voltest parameter\n");
      voltestboolean = 0x12345678;
      i++;

    } else if (strcmp(argv[i], "-dmawait") == 0) {
      if ((i+1) >= argc) {
	fprintf(stderr, "parameter missing for -dmawait\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      uade_dmawait = atoi(argv[i+1]);
      fprintf(stderr, "setting dmawait = %d rasterlines\n", uade_dmawait);
      uade_dmawait += 0x12340000;
      i += 2;

    } else if (strcmp(argv[i], "-ntsc") == 0) {
      uade_song.use_ntsc = 1;
      i++;

    } else if (strcmp(argv[i], "-noend") == 0) {
      uade_song.song_end_possible = 0;
      i++;

    } else if (strcmp(argv[i], "-sub") == 0) {
      if ((i+1) >= argc) {
	fprintf(stderr, "parameter missing for -sub\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      uade_song.subsong = atoi(argv[i+1]);
      if ((*argv[i+1] != '+') && (*argv[i+1] != '-')) { 
	uade_song.set_subsong = 1;
      } else {
	uade_song.set_subsong = 2;
	fprintf(stderr, "-sub parameter parsed, setting subsong %d+min_subsong\n", uade_song.subsong);
      }
      i+=2;

    } else if (strcmp(argv[i], "-one") == 0) {
      uade_one_subsong = 1;
      i++;

    } else if (strcmp(argv[i], "--verbose") == 0) {
      i++;

    } else if (strcmp(argv[i], "--version") == 0) {
      fprintf(stdout,"UADE %s\n", uade_version());
      uade_exit(0);

#ifdef HAVE_UNIX_SHELL
    } else if (strcmp(argv[i], "-varpid") == 0) {
      uade_create_var_pid();
      i++;
#endif

    } else if (strcmp(argv[i], "-pan")==0) {
      if ((i+1) >= argc) {
	fprintf(stderr, "-pan parameter missing\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      uade_do_panning = 1;
      uade_pan_value = atof(argv[i+1]);
      i += 2;

    } else if (strcmp(argv[i], "-config") == 0) {
      i++;

    } else if (strcmp(argv[i], "-repeat") == 0) {
      uade_repeat = 1;
      i++;

    } else if (strcmp(argv[i], "--swap-bytes") == 0) {
      uade_swap_output_bytes = 1;
      i++;

    } else if (strcmp(argv[i], "--no-mc") == 0) {
      disable_modulechange = 1;
      i++;

    } else if (!strcmp(argv[i], "--timeout") || !strcmp(argv[i], "-t")) {
      if ((i+1) >= argc) {
	fprintf(stderr, "timeout parameter missing\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      uade_timeout = atoi(argv[i+1]);
      i += 2;

    } else if (!strcmp(argv[i], "-st")) {
      if ((i+1) < argc) {
	uade_subsong_timeout = atoi(argv[i+1]);
      } else {
	fprintf(stderr, "per subsong timeout parameter missing\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      i += 2;

    } else if (!strcmp(argv[i], "-sit")) {
      if ((i+1) < argc) {
	uade_silence_timeout = atoi(argv[i+1]);
      } else {
	fprintf(stderr, "silence timeout parameter missing\n");
	uade_print_help(1);
	uade_exit(-1);
      }
      i += 2;

    } else if (!strcmp(argv[i], "--debug") || !strcmp(argv[i], "-d")) {
      uade_debug = 1;
      i++;

    } else if (!strcmp(argv[i], "-sh")) {
      uade_speed_hack = 1;
      i++;

    } else if (!strcmp(argv[i], "--no-sh")) {
      uade_speed_hack = -1;
      i++;

    } else if (!strcmp(argv[i], "--interactive") || !strcmp(argv[i], "-i")) {
      uade_interactive = 1;
      i++;

    } else if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
      uade_print_help(0);
      uade_exit(0);

    } else {

      local.enqueue_song(argv[i]);

      i++;
    }
  }

  if (uade_debug) {
    activate_debugger();
  }

  if (uade_remote_slave) {

    memset(&remote, 0, sizeof(remote));
#ifdef HAVE_XMMS_SLAVE
    xmms_slave_functions(&remote);
    uade_remote = &remote;
#endif
    if (remote.setup) {
      remote.setup(slaveargs);
    }

  } else {
    /* set default timeout on command line (don't set this if timeout was
       given with -t) */
    if (uade_timeout == -1) {
      uade_timeout = 10 * 60;  /* 10 minutes (i know, a stupid value) */
    }

    /* if songlist is empty, use given playername (-P) if exists */
    if (local.song_list_empty()) {
      if (uade_song.playername[0] == 0) {
	uade_print_help(2);
	uade_exit(-1);
      }
      /* queue zero string as module, only player is used */
      local.enqueue_song("");
    }
    local.setup(uade_interrupt);
  }

  uade_reboot = 1;
}


static void uade_print_help(int problemcode) {
  switch (problemcode) {
  case 0:
    /* just for printing help */
    break;
  case 1:
    fprintf(stderr, "Illegal parameters for UADE\n\n");
    break;
  case 2:
    fprintf(stderr, "No songs given as parameters for UADE\n\n");
    break;
  default:
    fprintf(stderr, "Unknown error\n");
    break;
  }
  fprintf(stderr, "UADE usage:\n");
  fprintf(stderr, " %s [OPTIONS] [FILE1] [FILE2] ...\n\n", uadecmdlinename);

  fprintf(stderr, "options:\n");
  fprintf(stderr, " -M songname\tSets songname to be played\n");
  fprintf(stderr, " -P playername\tSets eagleplayer to be used (you can use this for custom songs)\n");
  fprintf(stderr, " -sub number\tSets first subsong number to be played\n");
  fprintf(stderr, " -force\t\tForces eagleplayer to play a given song if it's not recognized\n");
  fprintf(stderr, " -repeat\tLoop song list to be played from beginning after the end\n");
  fprintf(stderr, " -pan value\tSets panning value in the range of [0.00, 1.00] (mix channels)\n");
  fprintf(stderr, " -one\t\tPlay at most one subsong / file\n");
  fprintf(stderr, " -noend\t\tThe first song will be played indefinetely (probably)\n");
  fprintf(stderr, " -ntsc\t\tForce Amiga into NTSC mode (may affect timing)\n");
  fprintf(stderr, " -outpipe x\tDump all sound output to file descriptor x (x=1 => stdout)\n");
  fprintf(stderr, " -S scorename\tForces scorename file to be used as 'sound core' (experts only)\n");
  fprintf(stderr, " --swap-bytes\tSwaps output sample bytes (use only with 16-bit samples)\n");
  fprintf(stderr, " -t timeout\tSets the maximum number of seconds to play a (sub)song\n");
  fprintf(stderr, " -st timeout\tSwitch to next subsong after 'timeout' seconds\n");
  fprintf(stderr, " -sit timeout\tSwitch to next subsong after 'timeout' seconds silent\n");
}

/* CRAP. REWRITE */
static void check_name_extension(char *playername, char *modname, int maxl) {
  FILE *formatsfile;
  FILE *songfile;
  struct stat st;
   
  int filesize = 5122;
  int realfilesize =0;
  unsigned char buf[5122] = "";
  int status;
  char formatsfilename[PATH_MAX];
  char playerdir[PATH_MAX];
  int x;
  char a[PATH_MAX], b[PATH_MAX], pre[PATH_MAX], post[PATH_MAX];

  *playername = 0;

  pre[0] = 0;
  post[0] = 0;
  if (!uade_get_prefix(pre, modname, sizeof(pre))) {
    return;
  }
  if (!uade_get_prefix(post, modname, sizeof(post))) {
    return;
  }

  if (!uade_get_path(formatsfilename, UADE_PATH_FORMATSFILE, sizeof(formatsfilename))) {
    return;
  }
  formatsfile = fopen(formatsfilename, "rb");
  if (!formatsfile) {
    return;
  }

  songfile = fopen (modname, "rb");
  if (!songfile) {
    fprintf (stderr, "uade: can not open file %s\n", modname);
    goto closeffile;
  }
  if ((status = decrunch ( &songfile, modname ) < 0 )) {
    fprintf (stderr, "decrunching error...\n");
    fclose (songfile);
    goto closeffile;
  }

  fstat (fileno(songfile),&st);
  realfilesize=st.st_size;

  fread  (buf, 1, 5122, songfile);
  fclose (songfile);
  filemagic (buf, pre, post, realfilesize);

  /* get player directory */
  uade_get_path(playerdir, UADE_PATH_PLAYERDIR, sizeof(playerdir));

  for(;;) {
    x = fscanf(formatsfile, "%s", a);
    if (x==0 || x==EOF) break;

    /* check if this is a comment line (this skips comment lines) */
    if (a[0]=='#') {
      while (a[0]!='\n') {
	x = fscanf(formatsfile,"%c",&a[0]);
	if (x==0 || x==EOF) break;
      }
      continue;
    }

    if (strcasecmp(a,"formats")==0) {
      for (;;) {
	x = fscanf(formatsfile,"%s",a);
	if (x == 0 || x==EOF) break;
	if (strcasecmp("endformats",a)==0) break;

	/* check if this is a comment line (this skips comment lines) */
	if (a[0] == '#') {
	  while (a[0] != '\n') {
	    x = fscanf(formatsfile, "%c", &a[0]);
	    if (x==0 || x==EOF) break;
	  }
	  continue;
	}
	
	x = fscanf(formatsfile,"%s",b);
	if (x==0 || x==EOF) break;

	if (strcasecmp(pre,a)==0 || strcasecmp(post,a)==0) {
	  if (strcmp(b,"custom")==0) {
	    strlcpy(playername, modname, maxl);
	    modname[0] = 0;
	  } else {
	    /* Find out player directory path */
	    strlcpy(playername, playerdir, maxl);
	    strlcat(playername, b, maxl);
	  }
	  goto closeffile;
	}
      }
      break;
    }
  }
 closeffile:
  fclose(formatsfile);
}


/* this is called for each played song from newcpu.c/m68k_reset() */
void uade_prerun(void) {
  /* don't load anything under 0x1000 (exec at $1000) */
  int modnameaddr = 0x00400;
  int scoreaddr   = 0x01000;
  int userstack   = 0x08500;
  int superstack  = 0x08f00;
  int playeraddr  = 0x09000;
  int relocaddr   = 0x40000;   /* default, recalculate this */
  int modaddr     = 0x80000;   /* default, recalculate this */
  int len;
  FILE* file;
  uae_u8* memp;
  int bytesread;
  /* decrunch variables */
  FILE* file_cr;
  char *modfile;
  int status;
  char tempname[PATH_MAX], playername[PATH_MAX];

  /* IMPORTANT:
     It seems that certain players don't work totally reliably if memory
     contains trash from previous songs. To be certain that each song is
     played from the same initial state of emulator we clear the memory
     from 0x400 to 0x100000 each time a new song is played */
  uade_highmem = 0;
  while (uade_highmem < 0x200000) {
    if (!valid_address(0, uade_highmem + 0x10000))
      break;
    uade_highmem += 0x10000;
  }
  if (uade_highmem < 0x200000) {
    fprintf(stderr, "uade: warning: highmem < 0x200000 (0x%x)!\n", uade_highmem);
  }
  if (valid_address(0x400, 0x100000-0x400)) {
    memset(get_real_address(0x400), 0, 0x100000-0x400);
  } else {
    fprintf(stderr, "uade: prerun: invalid memory range for clearing\n");
  }

  if (uade_remote_slave) {

    if (!remote.get_next) {
      fprintf(stderr, "uade: fatal error: no slave get next\n");
    }
    remote.get_next(&uade_song);

  } else {

  takenextsong:

    /* if no more songs in the play queue, quit cmdline tool */
    if (local.song_list_empty()) {
      fprintf(stderr,"uade: no more songs to play\n");
      uade_exit(0);
    }

    local.get_next(tempname, sizeof(tempname));
    strlcpy(uade_song.modulename, tempname, sizeof(uade_song.modulename));

    /* uade can be forced to use 'uade_song.playername' initialized with
       -P switch for all played songs. */
    if (!playerforced) {
      check_name_extension(playername, uade_song.modulename, sizeof(playername));
      /* playername is empty string if no playername was found */
      if (playername[0]) {
	strlcpy(uade_song.playername, playername, sizeof(uade_song.playername));
      } else {
	fprintf(stderr,"uade: file %s: unknown format\n", tempname);
      }
    }

    if (uade_repeat) {
      local.enqueue_song(tempname);
    }

    if (uade_song.playername[0] == 0)
      goto takenextsong;

    uade_cur_subsong = uade_min_subsong = uade_max_subsong = 0;
  }

  uade_set_automatic_song_end(uade_song.song_end_possible);

  if (uade_execdebugboolean)
    uade_put_long(SCORE_EXEC_DEBUG, 0x12345678);
  else
    uade_put_long(SCORE_EXEC_DEBUG, 0);

  uade_put_long(SCORE_VOLUME_TEST, voltestboolean);
  uade_put_long(SCORE_DMA_WAIT, uade_dmawait);
  uade_put_long(SCORE_MODULECHANGE, disable_modulechange);

  /* open player */
  file = fopen(uade_song.playername, "rb");
  if (file) {
    if ((status = decrunch (&file, uade_song.playername)) < 0) {
      fprintf (stderr, "decrunching error...");
      fclose(file); file = 0;
      goto setupfailure;
    } else {
      bytesread = uade_safe_load(playeraddr, file, uade_highmem - playeraddr);
      fclose(file); file = 0;
      fprintf(stderr, "uade: player '%s' (%d bytes)\n", uade_song.playername, bytesread);
      /* set player executable address for relocator */
      uade_put_long(SCORE_PLAYER_ADDR, playeraddr);
      memp = get_real_address(playeraddr);
      len = uade_calc_reloc_size((unsigned int *) memp);
      if (len) {
	relocaddr  = ((playeraddr + bytesread) & 0x7FFFF000) + 0x4000;
	/* + 0x4000 for hippel coso (wasseremu) */
	modaddr = ((relocaddr + len) & 0x7FFFF000) + 0x2000;
      } else {
	fprintf(stderr, "uade: problem with reloc calculation\n");
	if (!uade_remote_slave) {
 	  fprintf(stderr, "uade: skipping to next song\n");
	  goto takenextsong;
	}
      }
    }
  } else {
    fprintf(stderr,"uade: couldn't load player %s\n", uade_song.playername);
  setupfailure:
    if (uade_remote_slave) {
      remote.song_init_failure();
      return;
    }
    uade_exit(-1);
  }

  /* relocaddr = 0x40000;
     modaddr = 0x80000;
     fprintf(stderr, "relocaddr\t0x%x\nmodaddr\t\t0x%x\n", relocaddr, modaddr);
  */

  if (modaddr <= relocaddr) {
    /* this is very bad because sound core memory allocation will fail */
    fprintf(stderr, "uade: this is very bad: modaddr <= relocaddr: 0x%x <= 0x%x\n", modaddr, relocaddr);
  }

  uade_put_long(SCORE_RELOC_ADDR, relocaddr);  /*address for relocated player*/
  uade_put_long(SCORE_MODULE_ADDR, modaddr);   /* set module address */
  uade_put_long(SCORE_MODULE_LEN, 0);          /* set module size to zero */
  uade_put_long(SCORE_MODULE_NAME_ADDR, 0);    /* mod name address pointer */

  /* load the module if available */
  if (uade_song.modulename[0]) {
    if ((file = fopen(uade_song.modulename, "rb"))) {
      if ((status = decrunch (&file, uade_song.modulename)) < 0) {
	fprintf (stderr, "uade: decrunching error...");
	fclose(file); file = 0;
	goto setupfailure;
      } else {
	bytesread = uade_safe_load(modaddr, file, uade_highmem - modaddr);
	fclose(file); file = 0;
	fprintf(stderr, "uade: module '%s' (%d bytes)\n", uade_song.modulename, bytesread);
	uade_put_long(SCORE_MODULE_LEN, bytesread);
	strlcpy(get_real_address(modnameaddr), uade_song.modulename, 256);
	uade_put_long(SCORE_MODULE_NAME_ADDR, modnameaddr);
      }
    } else {
      goto setupfailure;
    }
  } else {
    memp = get_real_address(modnameaddr);
    strlcpy(memp, uade_song.playername, 256);
    uade_put_long(SCORE_MODULE_NAME_ADDR, modnameaddr);
  }

  /* load sound core (score) */
  if ((file = fopen(uade_song.scorename, "rb"))) {
    bytesread = uade_safe_load(scoreaddr, file, uade_highmem - scoreaddr);
    fclose(file); file = 0;
  } else {
    goto setupfailure;
  }

  m68k_setpc(scoreaddr);
  /* override bit for sound format checking */
  uade_put_long(SCORE_FORCE, uade_song.force_by_default);
  /* setsubsong */
  uade_put_long(SCORE_SET_SUBSONG, uade_song.set_subsong);
  uade_put_long(SCORE_SUBSONG, uade_song.subsong);
  /* set ntscbit correctly */
  uade_set_ntsc(uade_song.use_ntsc);
  /* pause bits (don't care!), for debugging purposes only */
  uade_put_long(SCORE_PREPAUSE, 0);
  uade_put_long(SCORE_POSTPAUSE, 0);
  /* set user and supervisor stack pointers */
  uade_put_long(SCORE_USER_STACK, userstack);
  uade_put_long(SCORE_SUPER_STACK, superstack);
  /* no message for score */
  uade_put_long(SCORE_OUTPUT_MSG, 0);
  if ( (userstack - (scoreaddr + bytesread)) < 0x1000) {
    fprintf(stderr, "uade: stack over run warning!\n");
  }

  uade_song.set_subsong = 0;

  if (!playerforced)
    uade_song.playername[0] = 0;

  uade_song.modulename[0] = 0;
  score_playername[0] = 0;
  score_modulename[0] = 0;

  if (uade_remote_slave) {
    remote.post_init();
  }

  uade_reset_counters();

  uade_flush_sound();

  /* note that uade_speed_hack can be negative (meaning that uade never uses
     speed hack, even if it's requested by the amiga player)! */
  uade_time_critical = 0;
  if (uade_speed_hack > 0) {
    uade_time_critical = 1;
  }

  uade_slave_end_reported = 0;

  uade_reboot = 0;
}

static void uade_put_long(int addr, int val) {
  uae_u8 *p;
  p = get_real_address(addr+0);
  *p = (uae_u8) (val>>24 & 0xff);
  p = get_real_address(addr+1);
  *p = (uae_u8) (val>>16 & 0xff);
  p = get_real_address(addr+2);
  *p = (uae_u8) (val>>8 & 0xff);
  p = get_real_address(addr+3);
  *p = (uae_u8) (val & 0xff);
}

static int uade_get_long(int addr) {
  uae_u8* ptr;
  int x = 0;
  ptr = get_real_address(addr);
  x += (*ptr) << 24;
  ptr = get_real_address(addr+1);
  x += (*ptr) << 16;
  ptr = get_real_address(addr+2);
  x += (*ptr) << 8;
  ptr = get_real_address(addr+3);
  x += (int) (*ptr);
  return x;
}

static int uade_safe_load(int dst, FILE *file, int maxlen) {
  const int bufsize = 4096;
  char buf[bufsize];
  int nbytes, len, off;
  len = bufsize;
  off = 0;
  if (maxlen <= 0)
    return 0;
  while (maxlen > 0) {
    if (maxlen < bufsize)
      len = maxlen;
    nbytes = fread(buf, 1, len, file);
    if (!nbytes)
      break;
    if (!valid_address(dst + off, nbytes)) {
      fprintf(stderr, "uade: illegal load range [%x,%x)\n", dst + off, dst + off + nbytes);
      break;
    }
    memcpy(get_real_address(dst + off), buf, nbytes);
    off += nbytes;
    maxlen -= nbytes;
  }
  return off;
}

static void uade_safe_get_string(char *dst, int src, int maxlen) {
  int i = 0;
  while(1) {
    if (i >= maxlen) {
      fprintf(stderr, "uade: safe_get_string: i over maxlen\n");
      break;
    }
    if (!valid_address(src + i, 1)) {
      fprintf(stderr, "uade: safe_get_string: invalid memory range\n");
      break;
    }
    dst[i] = *((char *) get_real_address(src + i));
    i++;
  }
  if (maxlen > 0) {
    if (i < maxlen) {
      dst[i] = 0;
    } else { 
      fprintf(stderr, "uade: safe_get_string: null termination warning!\n");
      dst[maxlen-1] = 0;
    }
  }
}

/* if kill_it is zero, uade may switch to next subsong. if kill_it is non-zero
   uade will always switch to next song (if any) */
void uade_song_end(char *reason, int kill_it) {

  if (uade_remote_slave) {
    if (uade_slave_end_reported) {
      return;
    }
  } else {
    if (uade_reboot) {
      return;
    }
  }

  fprintf(stderr, "uade: song end (%s)\n", reason);

  uade_reset_counters();

  if (uade_remote_slave) {
    uade_slave_end_reported = 1;
    remote.report(UADE_SONG_END, 0);

  } else {
    if (!kill_it) {
      if (uade_max_subsong && !uade_one_subsong) {
	if (uade_cur_subsong < uade_max_subsong) {
	  uade_cur_subsong++;
	  uade_change_subsong(uade_cur_subsong);
	  return;
	}
      }
    }
    uade_reboot = 1;
  }
}

void uade_get_amiga_message(void) {
  uae_u8 *ptr;
  uae_u8 *nameptr;
  FILE *file;
  int x;
  int mins, maxs, curs;
  int status;
  int src, dst, off, len;

  /* get input message type */
  x = uade_get_long(SCORE_INPUT_MSG);

  switch (x) {
  case UADE_SONG_END:
    uade_song_end("player", 0);
    break;

  case UADE_SUBSINFO:
    mins = uade_get_long(SCORE_MIN_SUBSONG);
    maxs = uade_get_long(SCORE_MAX_SUBSONG);
    curs = uade_get_long(SCORE_CUR_SUBSONG);
    fprintf(stderr, "uade: subsong info: minimum: %d maximum: %d current: %d\n", mins, maxs, curs);

    if (uade_remote_slave) {
      int subs[3] = {mins, maxs, curs};
      remote.report(UADE_SUBSINFO, subs);

    } else {
      uade_min_subsong = mins;
      uade_max_subsong = maxs;
      uade_cur_subsong = curs;
    }
    break;
  case UADE_PLAYERNAME:
    strlcpy(score_playername, get_real_address(0x204), sizeof(score_playername));
    fprintf(stderr,"uade: playername: %s\n", score_playername);
    if (uade_remote_slave) {
      remote.report(UADE_PLAYERNAME, score_playername);
    }
    break;

  case UADE_MODULENAME:
    strlcpy(score_modulename, get_real_address(0x204), sizeof(score_modulename));
    fprintf(stderr,"uade: modulename: %s\n", score_modulename);
    if (uade_remote_slave) {
      remote.report(UADE_MODULENAME, score_modulename);
    }
    break;
  case UADE_FORMATNAME:
    strlcpy(score_formatname, get_real_address(0x204), sizeof(score_formatname));
    fprintf(stderr,"uade: formatname: %s\n", score_formatname);
    if (uade_remote_slave) {
      remote.report(UADE_FORMATNAME, score_formatname);
    }
    break;
  case UADE_GENERALMSG:
    fprintf(stderr,"uade: general message: %s\n", get_real_address(0x204));
    break;
  case UADE_CHECKERROR:
    uade_song_end("module check failed", 1);
    break;
  case UADE_SCORECRASH:
    if (uade_debug) {
      fprintf(stderr,"uade: failure: score crashed\n");
      activate_debugger();
      break;
    }
    uade_song_end("score crashed", 1);
    break;
  case UADE_SCOREDEAD:
     if (uade_debug) {
      fprintf(stderr,"uade: score is dead\n"); 
      activate_debugger();
      break;
    }
     uade_song_end("score is dead", 1);
    break;
  case UADE_LOADFILE:
    /* load a file named at 0x204 (name pointer) to address pointed by
       0x208 and insert the length to 0x20C */
    src = uade_get_long(0x204);
    if (!valid_address(src, 1)) {
      fprintf(stderr, "uade: load: name in invalid address range\n");
      break;
    }
    nameptr = get_real_address(src);
    if ((file = uade_open_amiga_file(nameptr))) {
      if ((status = decrunch (&file, nameptr)) < 0) {
	fprintf (stderr, "uade: load: decrunching error\n");
	fclose(file);
	break;
      } else {
	dst = uade_get_long(0x208);
	len = uade_safe_load(dst, file, uade_highmem - dst);
	fclose(file); file = 0;
	uade_put_long(0x20C, len);
	/* fprintf(stderr, "uade: load: %s: ptr = 0x%x size = 0x%x\n", nameptr, dst, len); */
      }
    }
    break;
  case UADE_READ:
    src = uade_get_long(0x204);
    if (!valid_address(src, 1)) {
      fprintf(stderr, "uade: read: name in invalid address range\n");
      break;
    }
    nameptr = get_real_address(src);
    dst = uade_get_long(0x208);
    off = uade_get_long(0x20C);
    len = uade_get_long(0x210);
    /* fprintf(stderr,"uade: read: '%s' dst = 0x%x off = 0x%x len = 0x%x\n", nameptr, dst, off, len); */
    if ((file = uade_open_amiga_file(nameptr))) {
      fseek(file, off, SEEK_SET);
      x = uade_safe_load(dst, file, len);
      fclose(file); file = 0;
      uade_put_long(0x214, x);
    } else {
      fprintf(stderr, "uade: read: error when reading '%s'\n", nameptr);
      uade_put_long(0x214, 0);
    }
    break;
  case UADE_FILESIZE:
    src = uade_get_long(0x204);
    if (!valid_address(src, 1)) {
      fprintf(stderr, "uade: filesize: name in invalid address range\n");
      break;
    }
    nameptr = get_real_address(src);
    if ((file = uade_open_amiga_file(nameptr))) {
      fseek(file, 0, SEEK_END);
      len = ftell(file);
      fclose(file); file = 0;
      /* fprintf(stderr, "uade: size: 0x%x '%s'\n", len, nameptr); */
      uade_put_long(0x208, len);
      uade_put_long(0x20C, -1);
    } else {
      fprintf(stderr, "uade: can't get file size for '%s'\n", nameptr);
      uade_put_long(0x208, 0);
      uade_put_long(0x20C, 0);
    }
    break;
  case UADE_TIME_CRITICAL:
    uade_time_critical = uade_get_long(0x204) ? 1 : 0;
    if (uade_speed_hack < 0) {
      /* a negative value forbids use of speed hack */
      uade_time_critical = 0;
    }
    break;

  default:
    fprintf(stderr,"uade: unknown message from score ($%x)\n",x);
    break;
  }
}


void uade_change_subsong(int subs) {
  uade_put_long(SCORE_SUBSONG, subs);
  uade_send_amiga_message(UADE_SETSUBSONG);
  uade_flush_sound();
}

void uade_set_ntsc(int usentsc) {
  uade_put_long(SCORE_NTSC, usentsc);
}

void uade_set_automatic_song_end(int song_end_possible) {
  uade_put_long(SCORE_HAVE_SONGEND, song_end_possible);
}

void uade_send_amiga_message(int msgtype) {
  uade_put_long(SCORE_OUTPUT_MSG, msgtype);
}

static void uade_flush_sound(void) {
  if (uade_remote_slave) {
    remote.flush();
  }
  flush_sound();
}


static int uade_calc_reloc_size(unsigned int *src) {
  unsigned int offset;
  int i, nhunks;
  if (sizeof(unsigned int) != 4) {
    fprintf(stderr, "uade: reloc size problem: unsigned int != 4\n");
    fprintf(stderr, "uade: make reloc size calculator more portable\n");
    return 0;
  }
  if (ntohl(*src) != 0x03f3) return 0;
  src++;
  if (ntohl(*src)) return 0;
  src++;
  nhunks = ntohl(*src); src++; /* take number of hunks */
  src += 2;                    /* skip first & last hunk indexes */
  offset = 0;
  for(i = 0; i < nhunks; i++) {
    offset += 4 * (ntohl(*src) & 0x3FFFFFFF);
    src++;
  }
  return ((int) offset);
}


void uade_swap_buffer_bytes(void *data, int bytes) {
  uae_u8 *buf = (uae_u8 *) data;
  uae_u8 sample;
  int i;
  /* even number bytes */
  if ((bytes % 2) == 1)
    bytes--;
  for(i=0; i<bytes; i+=2) {
    sample = buf[i+0];
    buf[i+0] = buf[i+1];
    buf[i+1] = sample;
  }
}

void uade_reset_counters(void) {

  uade_zero_sample_count = 0;    /* only useful in non-slavemode */

  uade_vsync_counter = 0;
  uade_audxdat_counter = 0;
  uade_audxvol_counter = 0;
  uade_audxlch_counter = 0;
}

void uade_test_sound_block(void *buf, int size) {
  int i, s, exceptioncounter, bytes;

  if (uade_silence_timeout <= 0)
    return;

  if (currprefs.sound_bits == 16) {

    uae_s16 *sm = (uae_s16 *) buf;
    exceptioncounter = 0;

    for (i=0; i < (size/2); i++) {
      s = sm[i] >= 0 ? sm[i] : -sm[i];
      if (s >= (32767*1/100)) {
	exceptioncounter++;
	if (exceptioncounter > (size/2/100)) {
	  uade_zero_sample_count = 0;
	  break;
	}
      }
    }
    if (i == (size/2)) {
      uade_zero_sample_count += size;
    }

  } else {

    uae_s8 *sm = (uae_s8 *) buf;
    exceptioncounter = 0;

    for (i=0; i < size; i++) {
      s = sm[i] >= 0 ? sm[i] : -sm[i];
      if (s >= (127*2/100)) {
	exceptioncounter++;
	if (exceptioncounter > (size/100)) {
	  uade_zero_sample_count = 0;
	  break;
	}
      }
    }
    if (i == size) {
      uade_zero_sample_count += size;
    }
  }

  bytes = uade_silence_timeout * currprefs.sound_bits/8 * currprefs.sound_freq;
  if (currprefs.stereo) {
    bytes *= 2;
  }
  if (uade_zero_sample_count >= bytes) {
    char reason[256];
    sprintf(reason, "silence detected (%d seconds)", uade_silence_timeout);
    uade_song_end(reason, 0);
  }
}

static void uade_interaction(struct uade_command *cmd) {
  switch (cmd->type) {

  case UADE_NO_INTERACTION:
    uade_interactive = 0;
    fprintf(stderr, "uade: no interaction anymore\n");
    break;

  case UADE_REBOOT:
    uade_song_end("user request", 0);
    break;

  case UADE_SETSUBSONG:
    uade_change_subsong(*((int *) cmd->ret));
    break;

  default:
    fprintf(stderr, "uade: unkown command (%d)\n", cmd->type);
    break;
  }
}

void uade_vsync_handler(void) {

  uade_vsync_counter++;

  if (uade_timeout >= 0) {
    if ((uade_vsync_counter/50) >= uade_timeout) {
      char reason[256];
      sprintf(reason, "timeout %d seconds", uade_timeout);
      /* don't skip to next subsong even if available (kill == 1) */
      uade_song_end(reason, 1);
      return;
    }
  }

  if (uade_subsong_timeout >= 0) {
    if ((uade_vsync_counter/50) >= uade_subsong_timeout) {
      char reason[256];
      sprintf(reason, "per subsong timeout %d seconds", uade_subsong_timeout);
      /* skip to next subsong if available (kill == 0) */
      uade_song_end(reason, 0);
      return;
    }
  }

  if (uade_interactive) {
    if (local.interaction) {
      struct uade_command cmd;
      memset(&cmd, 0, sizeof(cmd));
      if (local.interaction(&cmd)) {
	uade_interaction(&cmd);
      }
      if (cmd.par) {
	free(cmd.par);
      }
      if (cmd.ret) {
	free(cmd.ret);
      }
    }
  }
}

void uade_audxdat(int nr, int v) {
}

void uade_audxlch(int nr, int v) {
  uade_audxlch_counter = uade_vsync_counter;
  /* fprintf(stderr, "aud%dlch 0x%.4x @ %d\n", nr, v, uade_audxlch_counter); */
}

void uade_audxvol(int nr, int v) {
  /* if zero is hit into a volume register, it is not noticed */
  if (v) {
    uade_audxvol_counter = uade_vsync_counter;
  }
  /* fprintf(stderr, "aud%dvol %d @ %d\n", nr, v, uade_audxvol_counter); */
}
