/* Copyright 2003 Heikki Orsila <heikki.orsila@ee.tut.fi>
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "../osdep/strl.c"

#include "uade.h"
#include "uade-os.h"
#include "xmms-slave.h"
#include "xmms-slave-msg.h"

struct uade_msgstruct *uade_struct;

static void uade_signalhandler(int);

static const int bufsize = sizeof(uade_struct->soundbuffer);


static int free_in_buffer(void) {
  int fs;
  int readoffset = uade_struct->sbuf_readoffset;
  /* Assumption is that write is always ahead or at the same position
     (this assumption means that if r=w then there is 32k free)
     cases are:
     1: r=w: free = 32k - 1 (else case)
     2: w>r: free = r + 32k - w - 1 (else case) (consistent with case 1)
     3: w<r: free = r - w - 1 (first case in if structure)
  */
  if (uade_struct->sbuf_writeoffset < readoffset) {
    /* case 3 */
    fs = readoffset - uade_struct->sbuf_writeoffset - 1;
  } else {
    /* case 1 and 2 */
    fs = bufsize + readoffset - uade_struct->sbuf_writeoffset - 1;
  }
  return fs;
}

void xmms_slave_write(void *sndbuffer, int bytes) {
  const int sleepunit = 10000;
  const int maxsleep = 1000000;
  int fs, sleepcounter, firstsize;

  if (!uade_struct->dontwritebit) {

    for (sleepcounter = 0; sleepcounter < maxsleep;) {
      fs = free_in_buffer();
      if (fs >= bytes)
	break;
      if (uade_struct->dontwritebit)
	break;
      uade_usleep(sleepunit);
      sleepcounter += sleepunit;
    }

    while (uade_struct->plugin_pause_boolean)
      uade_usleep(sleepunit);

    /* copy sndbuffer to uade shared memory sound buffer (take into
       account that the buffer is circular) */
    if((uade_struct->sbuf_writeoffset + bytes) > bufsize) {
      firstsize = bufsize - uade_struct->sbuf_writeoffset;
      memcpy(uade_struct->soundbuffer + uade_struct->sbuf_writeoffset, sndbuffer, firstsize);
      memcpy(uade_struct->soundbuffer, ((char *) sndbuffer) + firstsize, bytes - firstsize);
    } else {
      memcpy(uade_struct->soundbuffer + uade_struct->sbuf_writeoffset, sndbuffer, bytes);
    }
    /* increase write counter */
    uade_struct->sbuf_writeoffset = (uade_struct->sbuf_writeoffset + bytes) % bufsize;
  }
}

static UADE_SIGNALHANDLER(uade_signalhandler) {
  int msg = uade_struct->touaemsgtype;

  uade_struct->touaemsgtype = 0;

  switch(msg) {
  case UADE_REBOOT:
    /* reboot, switch to next song, ... */
    uade_reboot = 1;
    break;

  case UADE_SETSUBSONG:
    /* this is used to set new subsong during playback */
    uade_change_subsong(uade_struct->subsong);
    uade_struct->set_subsong = 0;
    break;

  case UADE_PLAYERNAME:
    /* this signal informs that (playername, modulename, scorename)
       has been set into uade_struct */
    break;

  case UADE_NTSC:
    /* switch between ntsc/pal mode */
    uade_set_ntsc(uade_struct->ntscbit);
    uade_send_amiga_message(UADE_NTSC);
    break;

  case UADE_SONG_END:
    /* this does not signel to end song playing. sets automatic song end
       on (1) or off (0) */
    uade_set_automatic_song_end(uade_struct->songendpossible);
    break;

  default:
    fprintf(stderr,"uade: signal handler: msg was not recognized (%d)\n", msg);
    break;
  }
}

/* called once when uade process is started */
int xmms_slave_setup(char **slaveargs) {
  char *mapname;
  pid_t father;

  if (!slaveargs) {
    fprintf(stderr, "uade: xmms_slave_setup(): slaveargs == 0!\n");
    uade_exit(-1);
  }
  mapname = slaveargs[0];

  uade_struct = (struct uade_msgstruct *) uade_mmap_file(mapname);
  if (!uade_struct) {
    fprintf(stderr, "uade.c/uade: couldn't mmap file\n");
    uade_exit(-1);
  }

  /* setup a signal handler for communication with xmms plugin */
  uade_create_signalhandler(uade_signalhandler, 1);

  /* leave a watchdog that checks if XMMS plugin dies (polls plugin
     task with a kill(pluginpid, 0) signal every 0.5 seconds). If XMMS
     plugin dies, watchdog kills UADE and itself */
  father = getpid();
  if (!fork()) {
    while (1) {
      uade_sleep(2);
      if (uade_struct->masterpid) {
	if (kill(uade_struct->masterpid, 0) < 0) {
	  break;
	}
      }
      if (kill(father, 0) < 0) {
	fprintf(stderr, "uade: interesting... my father has died. i will go too...\n");
	break;
      }
    }
    /* remove the shared memory file that is used by uade and xmms plugin */
    (void) remove(mapname);
    kill(father, SIGKILL);
    uade_exit(-1);
  }

  uade_struct->min_subsong = 0;
  uade_struct->max_subsong = 0;

  return 1;
}

void xmms_slave_get_next(struct uade_song *song) {
  /* RACE CONDITION BEGINS (PENALTY OF LOSING IS 2 SECONDS) */
  /* uade executable is ready to receive commands */
  uade_struct->uade_inited_boolean = 1;
  /* uade_sleep() is interrupted by a SIGHUP from XMMS plugin */
  while (uade_struct->loadnewsongboolean == 0) {
    uade_sleep(2);
  }
  /* RACE CONDITION ENDS */

  strlcpy(song->playername, uade_struct->playername, sizeof(song->playername));
  strlcpy(song->modulename, uade_struct->modulename, sizeof(song->modulename));
  strlcpy(song->scorename, uade_struct->scorename, sizeof(song->scorename));

  uade_struct->loadnewsongboolean = 0;
    
  song->force_by_default = uade_struct->force_by_default;
  if (uade_struct->set_subsong) {
    uade_struct->set_subsong = 0;
    song->set_subsong = 1;
    song->subsong = uade_struct->subsong;
  }
  uade_struct->min_subsong = uade_struct->max_subsong = 0;
  song->song_end_possible = uade_struct->songendpossible;

  song->use_ntsc = uade_struct->ntscbit;
}

void xmms_slave_song_init_failure(void) {
  uade_reboot = 0;
  uade_struct->song_end = 1;
}

void xmms_slave_post_init(void) {
  uade_struct->score_playername[0] = 0;
  uade_struct->score_playerauthor[0] = 0;
  uade_struct->score_modulename[0] = 0;
  uade_struct->score_formatname[0] = 0;
}

void xmms_slave_report(int type, void *data) {
  char *s;
  int *subs;

  switch (type) {

  case UADE_SONG_END:
    uade_struct->song_end = 1;
    break;

  case UADE_SUBSINFO:
    subs = (int *) data;
    uade_struct->min_subsong = subs[0];
    uade_struct->max_subsong = subs[1];
    uade_struct->subsong = subs[2];
    break;

  case UADE_PLAYERNAME:
    strlcpy(uade_struct->score_playername, (char *) data, sizeof(uade_struct->score_playername));
    break;

  case UADE_FORMATNAME:
    strlcpy(uade_struct->score_formatname, (char *) data, sizeof(uade_struct->score_formatname));
    break;

  case UADE_MODULENAME:
    strlcpy(uade_struct->score_modulename, (char *) data, sizeof(uade_struct->score_modulename));
    break;

  default:
    fprintf(stderr, "unknown slave report\n");
    exit(-1);
  }
}

void xmms_slave_flush(void) {
  memset(uade_struct->soundbuffer, 0, sizeof(uade_struct->soundbuffer));
}

void xmms_slave_functions(struct uade_remote *remote) {
  remote->setup = xmms_slave_setup;
  remote->get_next = xmms_slave_get_next;
  remote->song_init_failure = xmms_slave_song_init_failure;
  remote->post_init = xmms_slave_post_init;
  remote->report = xmms_slave_report;
  remote->flush = xmms_slave_flush;
  remote->write = xmms_slave_write;
}
